@@ -18,6 +18,13 @@
 #include <string.h>
 #include <stdarg.h>
 
+/* Win32 compatibility */
+#if defined(_WIN32)
+# define vsnprintf _vsnprintf
+# define __func__ __FUNCTION__
+# define ZLIB_WINAPI 1
+#endif
+
 #if defined(HAVE_ZLIB)
 # include <zlib.h>
 #endif
@@ -452,6 +459,7 @@ static void _negotiate(telnet_t *telnet, unsigned char telopt) {
 			break;
 		case Q_WANTNO_OP:
 			_set_rfc1143(telnet, telopt, Q_WANTYES, Q_HIM(q));
+			_send_negotiate(telnet, TELNET_WILL, telopt);
 			NEGOTIATE_EVENT(telnet, TELNET_EV_WILL, telopt);
 			break;
 		case Q_WANTYES:
@@ -475,7 +483,7 @@ static void _negotiate(telnet_t *telnet, unsigned char telopt) {
  * value strings are NUL-terminated, all while fitting inside
  * of the original buffer.
  */
-static int _environ(telnet_t *telnet, unsigned char type,
+static int _environ_telnet(telnet_t *telnet, unsigned char type,
 		char* buffer, size_t size) {
 	telnet_event_t ev;
 	struct telnet_environ_t *values = 0;
@@ -617,7 +625,7 @@ static int _environ(telnet_t *telnet, unsigned char type,
 }
 
 /* process an MSSP subnegotiation buffer */
-static int _mssp(telnet_t *telnet, char* buffer, size_t size) {
+static int _mssp_telnet(telnet_t *telnet, char* buffer, size_t size) {
 	telnet_event_t ev;
 	struct telnet_environ_t *values;
 	char *var = 0;
@@ -696,9 +704,9 @@ static int _mssp(telnet_t *telnet, char* buffer, size_t size) {
 }
 
 /* parse ZMP command subnegotiation buffers */
-static int _zmp(telnet_t *telnet, const char* buffer, size_t size) {
+static int _zmp_telnet(telnet_t *telnet, const char* buffer, size_t size) {
 	telnet_event_t ev;
-	const char **argv;
+	char **argv;
 	const char *c;
 	size_t i, argc;
 
@@ -714,7 +722,7 @@ static int _zmp(telnet_t *telnet, const char* buffer, size_t size) {
 		c += strlen(c) + 1;
 
 	/* allocate argument array, bail on error */
-	if ((argv = (const char **)calloc(argc, sizeof(char *))) == 0) {
+	if ((argv = (char **)calloc(argc, sizeof(char *))) == 0) {
 		_error(telnet, __LINE__, __func__, TELNET_ENOMEM, 0,
 				"calloc() failed: %s", strerror(errno));
 		return 0;
@@ -722,13 +730,13 @@ static int _zmp(telnet_t *telnet, const char* buffer, size_t size) {
 
 	/* populate argument array */
 	for (i = 0, c = buffer; i != argc; ++i) {
-		argv[i] = c;
+		argv[i] = (char *)c;
 		c += strlen(c) + 1;
 	}
 
 	/* invoke event with our arguments */
 	ev.type = TELNET_EV_ZMP;
-	ev.zmp.argv = argv;
+	ev.zmp.argv = (const char**)argv;
 	ev.zmp.argc = argc;
 	telnet->eh(telnet, &ev, telnet->ud);
 
@@ -738,7 +746,7 @@ static int _zmp(telnet_t *telnet, const char* buffer, size_t size) {
 }
 
 /* parse TERMINAL-TYPE command subnegotiation buffers */
-static int _ttype(telnet_t *telnet, const char* buffer, size_t size) {
+static int _ttype_telnet(telnet_t *telnet, const char* buffer, size_t size) {
 	telnet_event_t ev;
 
 	/* make sure request is not empty */
@@ -761,7 +769,7 @@ static int _ttype(telnet_t *telnet, const char* buffer, size_t size) {
 		char *name;
 
 		/* allocate space for name */
-		if ((name = (char *)malloc(size - 1)) == 0) {
+		if ((name = (char *)malloc(size)) == 0) {
 			_error(telnet, __LINE__, __func__, TELNET_ENOMEM, 0,
 					"malloc() failed: %s", strerror(errno));
 			return 0;
@@ -820,15 +828,15 @@ static int _subnegotiate(telnet_t *telnet) {
 
 	/* specially handled subnegotiation telopt types */
 	case TELNET_TELOPT_ZMP:
-		return _zmp(telnet, telnet->buffer, telnet->buffer_pos);
+		return _zmp_telnet(telnet, telnet->buffer, telnet->buffer_pos);
 	case TELNET_TELOPT_TTYPE:
-		return _ttype(telnet, telnet->buffer, telnet->buffer_pos);
+		return _ttype_telnet(telnet, telnet->buffer, telnet->buffer_pos);
 	case TELNET_TELOPT_ENVIRON:
 	case TELNET_TELOPT_NEW_ENVIRON:
-		return _environ(telnet, telnet->sb_telopt, telnet->buffer,
+		return _environ_telnet(telnet, telnet->sb_telopt, telnet->buffer,
 				telnet->buffer_pos);
 	case TELNET_TELOPT_MSSP:
-		return _mssp(telnet, telnet->buffer, telnet->buffer_pos);
+		return _mssp_telnet(telnet, telnet->buffer, telnet->buffer_pos);
 	default:
 		return 0;
 	}
@@ -1339,16 +1347,14 @@ void telnet_begin_compress2(telnet_t *telnet) {
 }
 
 /* send formatted data with \r and \n translation in addition to IAC IAC */
-int telnet_printf(telnet_t *telnet, const char *fmt, ...) {
+int telnet_vprintf(telnet_t *telnet, const char *fmt, va_list va) {
     static const char CRLF[] = { '\r', '\n' };
     static const char CRNUL[] = { '\r', '\0' };
 	char buffer[1024];
 	char *output = buffer;
-	va_list va;
 	int rs, i, l;
 
 	/* format */
-	va_start(va, fmt);
 	rs = vsnprintf(buffer, sizeof(buffer), fmt, va);
 	if (rs >= sizeof(buffer)) {
 		output = (char*)malloc(rs + 1);
@@ -1359,7 +1365,6 @@ int telnet_printf(telnet_t *telnet, const char *fmt, ...) {
 		}
 		rs = vsnprintf(output, rs + 1, fmt, va);
 	}
-	va_end(va);
 
 	/* send */
 	for (l = i = 0; i != rs; ++i) {
@@ -1396,15 +1401,25 @@ int telnet_printf(telnet_t *telnet, const char *fmt, ...) {
 	return rs;
 }
 
-/* send formatted data through telnet_send */
-int telnet_raw_printf(telnet_t *telnet, const char *fmt, ...) {
-	char buffer[1024];
-	char *output = buffer;
+/* see telnet_vprintf */
+int telnet_printf(telnet_t *telnet, const char *fmt, ...) {
 	va_list va;
 	int rs;
 
-	/* format; allocate more space if necessary */
 	va_start(va, fmt);
+	rs = telnet_vprintf(telnet, fmt, va);
+	va_end(va);
+
+	return rs;
+}
+
+/* send formatted data through telnet_send */
+int telnet_raw_vprintf(telnet_t *telnet, const char *fmt, va_list va) {
+	char buffer[1024];
+	char *output = buffer;
+	int rs;
+
+	/* format; allocate more space if necessary */
 	rs = vsnprintf(buffer, sizeof(buffer), fmt, va);
 	if (rs >= sizeof(buffer)) {
 		output = (char*)malloc(rs + 1);
@@ -1415,7 +1430,6 @@ int telnet_raw_printf(telnet_t *telnet, const char *fmt, ...) {
 		}
 		rs = vsnprintf(output, rs + 1, fmt, va);
 	}
-	va_end(va);
 
 	/* send out the formatted data */
 	telnet_send(telnet, output, rs);
@@ -1428,6 +1442,18 @@ int telnet_raw_printf(telnet_t *telnet, const char *fmt, ...) {
 	return rs;
 }
 
+/* see telnet_raw_vprintf */
+int telnet_raw_printf(telnet_t *telnet, const char *fmt, ...) {
+	va_list va;
+	int rs;
+
+	va_start(va, fmt);
+	rs = telnet_raw_vprintf(telnet, fmt, va);
+	va_end(va);
+
+	return rs;
+}
+
 /* begin NEW-ENVIRON subnegotation */
 void telnet_begin_newenviron(telnet_t *telnet, unsigned char cmd) {
 	telnet_begin_sb(telnet, TELNET_TELOPT_NEW_ENVIRON);
@@ -1475,23 +1501,29 @@ void telnet_send_zmp(telnet_t *telnet, size_t argc, const char **argv) {
 }
 
 /* send ZMP data using varargs  */
-void telnet_send_zmpv(telnet_t *telnet, ...) {
-	va_list va;
+void telnet_send_vzmpv(telnet_t *telnet, va_list va) {
 	const char* arg;
 
 	/* ZMP header */
 	telnet_begin_sb(telnet, TELNET_TELOPT_ZMP);
 
 	/* send out each argument, including trailing NUL byte */
-	va_start(va, telnet);
 	while ((arg = va_arg(va, const char *)) != 0)
 		telnet_zmp_arg(telnet, arg);
-	va_end(va);
 
 	/* ZMP footer */
 	telnet_finish_zmp(telnet);
 }
 
+/* see telnet_send_vzmpv */
+void telnet_send_zmpv(telnet_t *telnet, ...) {
+	va_list va;
+
+	va_start(va, telnet);
+	telnet_send_vzmpv(telnet, va);
+	va_end(va);
+}
+
 /* begin a ZMP command */
 void telnet_begin_zmp(telnet_t *telnet, const char *cmd) {
 	telnet_begin_sb(telnet, TELNET_TELOPT_ZMP);
